var toStringFunction = Function.prototype.toString;
var create = Object.create;
var LegacyCache = /** @class */ (function () {
    function LegacyCache() {
        this._keys = [];
        this._values = [];
    }
    LegacyCache.prototype.has = function (key) {
        return !!~this._keys.indexOf(key);
    };
    LegacyCache.prototype.get = function (key) {
        return this._values[this._keys.indexOf(key)];
    };
    LegacyCache.prototype.set = function (key, value) {
        this._keys.push(key);
        this._values.push(value);
    };
    return LegacyCache;
}());
function createCacheLegacy() {
    return new LegacyCache();
}
function createCacheModern() {
    return new WeakMap();
}
/**
 * Get a new cache object to prevent circular references.
 */
var createCache = typeof WeakMap !== 'undefined' ? createCacheModern : createCacheLegacy;
/**
 * Get an empty version of the object with the same prototype it has.
 */
function getCleanClone(prototype) {
    if (!prototype) {
        return create(null);
    }
    var Constructor = prototype.constructor;
    if (Constructor === Object) {
        return prototype === Object.prototype ? {} : create(prototype);
    }
    if (~toStringFunction.call(Constructor).indexOf('[native code]')) {
        try {
            return new Constructor();
        }
        catch (_a) { }
    }
    return create(prototype);
}
function getRegExpFlagsLegacy(regExp) {
    var flags = '';
    if (regExp.global) {
        flags += 'g';
    }
    if (regExp.ignoreCase) {
        flags += 'i';
    }
    if (regExp.multiline) {
        flags += 'm';
    }
    if (regExp.unicode) {
        flags += 'u';
    }
    if (regExp.sticky) {
        flags += 'y';
    }
    return flags;
}
function getRegExpFlagsModern(regExp) {
    return regExp.flags;
}
/**
 * Get the flags to apply to the copied regexp.
 */
var getRegExpFlags = /test/g.flags === 'g' ? getRegExpFlagsModern : getRegExpFlagsLegacy;

var _a, _b;
var isArray = Array.isArray;
var defineProperty = Object.defineProperty, getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor, getOwnPropertyNames = Object.getOwnPropertyNames, getOwnPropertySymbols = Object.getOwnPropertySymbols, getPrototypeOf = Object.getPrototypeOf;
var _c = Object.prototype, hasOwnProperty = _c.hasOwnProperty, propertyIsEnumerable = _c.propertyIsEnumerable;
var ARRAY_BUFFER_OBJECT_CLASSES = (_a = {},
    _a['[object ArrayBuffer]'] = true,
    _a['[object Float32Array]'] = true,
    _a['[object Float64Array]'] = true,
    _a['[object Int8Array]'] = true,
    _a['[object Int16Array]'] = true,
    _a['[object Int32Array]'] = true,
    _a['[object Uint8Array]'] = true,
    _a['[object Uint8ClampedArray]'] = true,
    _a['[object Uint16Array]'] = true,
    _a['[object Uint32Array]'] = true,
    _a['[object Uint64Array]'] = true,
    _a);
var UNCOPIABLE_OBJECT_CLASSES = (_b = {},
    _b['[object Error]'] = true,
    _b['[object Promise]'] = true,
    _b['[object WeakMap]'] = true,
    _b['[object WeakSet]'] = true,
    _b);
var SUPPORTS_SYMBOL = typeof getOwnPropertySymbols === 'function';
function getStrictPropertiesModern(object) {
    return getOwnPropertyNames(object).concat(getOwnPropertySymbols(object));
}
var getStrictProperties = SUPPORTS_SYMBOL
    ? getStrictPropertiesModern
    : getOwnPropertyNames;
function copyArrayLoose(array, prototype, copier, cache) {
    var clone = new prototype.constructor();
    cache.set(array, clone);
    for (var index = 0, length_1 = array.length; index < length_1; ++index) {
        clone[index] = copier(array[index], cache);
    }
    return clone;
}
function copyObjectLooseLegacy(object, prototype, copier, cache) {
    var clone = getCleanClone(prototype);
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    for (var key in object) {
        if (hasOwnProperty.call(object, key)) {
            clone[key] = copier(object[key], cache);
        }
    }
    return clone;
}
function copyObjectLooseModern(object, prototype, copier, cache) {
    var clone = getCleanClone(prototype);
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    for (var key in object) {
        if (hasOwnProperty.call(object, key)) {
            clone[key] = copier(object[key], cache);
        }
    }
    var symbols = getOwnPropertySymbols(object);
    for (var index = 0, length_2 = symbols.length, symbol = void 0; index < length_2; ++index) {
        symbol = symbols[index];
        if (propertyIsEnumerable.call(object, symbol)) {
            clone[symbol] = copier(object[symbol], cache);
        }
    }
    return clone;
}
/**
 * Get a copy of the object based on loose rules, meaning all enumerable keys
 * and symbols are copied, but property descriptors are not considered.
 */
var copyObjectLoose = SUPPORTS_SYMBOL
    ? copyObjectLooseModern
    : copyObjectLooseLegacy;
/**
 * Get a copy of the object based on strict rules, meaning all keys and symbols
 * are copied based on the original property descriptors.
 */
function copyObjectStrict(object, prototype, copier, cache) {
    var clone = getCleanClone(prototype);
    // set in the cache immediately to be able to reuse the object recursively
    cache.set(object, clone);
    var properties = getStrictProperties(object);
    for (var index = 0, length_3 = properties.length, property = void 0, descriptor = void 0; index < length_3; ++index) {
        property = properties[index];
        if (property === 'callee' || property === 'caller') {
            continue;
        }
        descriptor = getOwnPropertyDescriptor(object, property);
        if (!descriptor) {
            // In extra edge cases where the property descriptor cannot be retrived, fall back to
            // the loose assignment.
            clone[property] = copier(object[property], cache);
            continue;
        }
        // Only clone the value if actually a value, not a getter / setter.
        if (!descriptor.get && !descriptor.set) {
            descriptor.value = copier(descriptor.value, cache);
        }
        try {
            defineProperty(clone, property, descriptor);
        }
        catch (error) {
            // Tee above can fail on node in edge cases, so fall back to the loose assignment.
            clone[property] = descriptor.value;
        }
    }
    return clone;
}
function copyMap(value, Constructor, copier, cache) {
    var clone = new Constructor();
    cache.set(value, clone);
    value.forEach(function (v, k) {
        clone.set(k, copier(v, cache));
    });
    return value;
}
function copyRegExp(value, Constructor) {
    var clone = new Constructor(value.source, getRegExpFlags(value));
    clone.lastIndex = value.lastIndex;
    return clone;
}
function copySet(value, Constructor, copier, cache) {
    var clone = new Constructor();
    cache.set(value, clone);
    value.forEach(function (v) {
        clone.add(copier(v, cache));
    });
    return value;
}
function createCopier(isStrict) {
    var copyArray = isStrict ? copyObjectStrict : copyArrayLoose;
    var copyObject = isStrict ? copyObjectStrict : copyObjectLoose;
    return function copier(value, cache) {
        if (!value || typeof value !== 'object') {
            return value;
        }
        if (cache.has(value)) {
            return cache.get(value);
        }
        var prototype = value.__proto__ || getPrototypeOf(value);
        var Constructor = prototype && prototype.constructor;
        // plain objects
        if (!Constructor || Constructor === Object) {
            return copyObject(value, prototype, copier, cache);
        }
        // arrays
        if (isArray(value)) {
            return copyArray(value, prototype, copier, cache);
        }
        var objectClass = toString.call(value);
        // dates
        if (objectClass === '[object Date]') {
            return new Constructor(value.getTime());
        }
        // regexps
        if (objectClass === '[object RegExp]') {
            return copyRegExp(value, Constructor);
        }
        // maps
        if (objectClass === '[object Map]') {
            return copyMap(value, Constructor, copier, cache);
        }
        // sets
        if (objectClass === '[object Set]') {
            return copySet(value, Constructor, copier, cache);
        }
        // blobs
        if (objectClass === '[object Blob]') {
            return value.slice(0, value.size, value.type);
        }
        // dataviews
        if (objectClass === '[object DataView]') {
            return new Constructor(value.buffer.slice(0));
        }
        // array buffers
        if (ARRAY_BUFFER_OBJECT_CLASSES[objectClass]) {
            return value.slice(0);
        }
        // if the value cannot / should not be copied deeply, return the reference
        if (
        // promise-like
        typeof value.then === 'function' ||
            // object classes which cannot be introspected for copy
            UNCOPIABLE_OBJECT_CLASSES[objectClass]) {
            return value;
        }
        // assume anything left is a custom constructor
        return copyObject(value, prototype, copier, cache);
    };
}

var copierStrict = createCopier(true);
var copierLoose = createCopier(false);
/**
 * Copy an value deeply as much as possible.
 */
function copy(value) {
    return copierLoose(value, createCache());
}
/**
 * Copy an value deeply as much as possible, where strict recreation of object properties
 * are maintained. All properties (including non-enumerable ones) are copied with their
 * original property descriptors on both objects and arrays.
 */
function copyStrict(value) {
    return copierStrict(value, createCache());
}

export { copyStrict, copy as default };
//# sourceMappingURL=index.mjs.map
